@@ -67,12 +67,23 @@
  * $Id$
  */
 
+/******************************************************************************\
+*
+\******************************************************************************/
+
+#include "jpc_t2cod.h"
+#include "jpc_cs.h"
+#include "jpc_math.h"
+
 #include "jasper/jas_math.h"
 #include "jasper/jas_malloc.h"
+#include "jasper/jas_debug.h"
 
-#include "jpc_cs.h"
-#include "jpc_t2cod.h"
-#include "jpc_math.h"
+#include <assert.h>
+
+/******************************************************************************\
+*
+\******************************************************************************/
 
 static int jpc_pi_nextlrcp(jpc_pi_t *pi);
 static int jpc_pi_nextrlcp(jpc_pi_t *pi);
@@ -80,9 +91,36 @@ static int jpc_pi_nextrpcl(jpc_pi_t *pi);
 static int jpc_pi_nextpcrl(jpc_pi_t *pi);
 static int jpc_pi_nextcprl(jpc_pi_t *pi);
 
+/******************************************************************************\
+*
+\******************************************************************************/
+
+static void jpc_pirlvl_destroy(jpc_pirlvl_t *rlvl)
+{
+	if (rlvl->prclyrnos) {
+		jas_free(rlvl->prclyrnos);
+	}
+}
+
+static void jpc_picomp_destroy(jpc_picomp_t *picomp)
+{
+	unsigned rlvlno;
+	jpc_pirlvl_t *pirlvl;
+	if (picomp->pirlvls) {
+		for (rlvlno = 0, pirlvl = picomp->pirlvls; rlvlno <
+		  picomp->numrlvls; ++rlvlno, ++pirlvl) {
+			jpc_pirlvl_destroy(pirlvl);
+		}
+		jas_free(picomp->pirlvls);
+	}
+}
+
+/******************************************************************************\
+*
+\******************************************************************************/
+
 int jpc_pi_next(jpc_pi_t *pi)
 {
-	jpc_pchg_t *pchg;
 	int ret;
 
 	for (;;) {
@@ -96,16 +134,19 @@ int jpc_pi_next(jpc_pi_t *pi)
 			pi->prcno = 0;
 			pi->lyrno = 0;
 			pi->prgvolfirst = true;
-			if (pi->pchgno < jpc_pchglist_numpchgs(pi->pchglist)) {
+			if (JAS_CAST(unsigned, pi->pchgno) <
+			  jpc_pchglist_numpchgs(pi->pchglist)) {
 				pi->pchg = jpc_pchglist_get(pi->pchglist, pi->pchgno);
-			} else if (pi->pchgno == jpc_pchglist_numpchgs(pi->pchglist)) {
+			} else if (JAS_CAST(unsigned, pi->pchgno) ==
+			  jpc_pchglist_numpchgs(pi->pchglist)) {
 				pi->pchg = &pi->defaultpchg;
 			} else {
+				JAS_LOGDEBUGF(10, "jpc_pi_next returning 1\n");
 				return 1;
 			}
 		}
 
-		pchg = pi->pchg;
+		const jpc_pchg_t *pchg = pi->pchg;
 		switch (pchg->prgord) {
 		case JPC_COD_LRCPPRG:
 			ret = jpc_pi_nextlrcp(pi);
@@ -137,10 +178,11 @@ int jpc_pi_next(jpc_pi_t *pi)
 
 static int jpc_pi_nextlrcp(register jpc_pi_t *pi)
 {
-	jpc_pchg_t *pchg;
-	int *prclyrno;
+	unsigned *prclyrno;
 
-	pchg = pi->pchg;
+	JAS_LOGDEBUGF(10, "jpc_pi_nextlrcp\n");
+
+	const jpc_pchg_t *pchg = pi->pchg;
 	if (!pi->prgvolfirst) {
 		prclyrno = &pi->pirlvl->prclyrnos[pi->prcno];
 		goto skip;
@@ -149,12 +191,12 @@ static int jpc_pi_nextlrcp(register jpc_pi_t *pi)
 	}
 
 	for (pi->lyrno = 0; pi->lyrno < pi->numlyrs && pi->lyrno <
-	  JAS_CAST(int, pchg->lyrnoend); ++pi->lyrno) {
+	  pchg->lyrnoend; ++pi->lyrno) {
 		for (pi->rlvlno = pchg->rlvlnostart; pi->rlvlno < pi->maxrlvls &&
 		  pi->rlvlno < pchg->rlvlnoend; ++pi->rlvlno) {
 			for (pi->compno = pchg->compnostart, pi->picomp =
 			  &pi->picomps[pi->compno]; pi->compno < pi->numcomps
-			  && pi->compno < JAS_CAST(int, pchg->compnoend); ++pi->compno,
+			  && pi->compno < pchg->compnoend; ++pi->compno,
 			  ++pi->picomp) {
 				if (pi->rlvlno >= pi->picomp->numrlvls) {
 					continue;
@@ -180,10 +222,10 @@ skip:
 
 static int jpc_pi_nextrlcp(register jpc_pi_t *pi)
 {
-	jpc_pchg_t *pchg;
-	int *prclyrno;
+	unsigned *prclyrno;
+	JAS_LOGDEBUGF(10, "jpc_pi_nextrlcp\n");
 
-	pchg = pi->pchg;
+	const jpc_pchg_t *pchg = pi->pchg;
 	if (!pi->prgvolfirst) {
 		assert(pi->prcno < pi->pirlvl->numprcs);
 		prclyrno = &pi->pirlvl->prclyrnos[pi->prcno];
@@ -195,10 +237,10 @@ static int jpc_pi_nextrlcp(register jpc_pi_t *pi)
 	for (pi->rlvlno = pchg->rlvlnostart; pi->rlvlno < pi->maxrlvls &&
 	  pi->rlvlno < pchg->rlvlnoend; ++pi->rlvlno) {
 		for (pi->lyrno = 0; pi->lyrno < pi->numlyrs && pi->lyrno <
-		  JAS_CAST(int, pchg->lyrnoend); ++pi->lyrno) {
+		  pchg->lyrnoend; ++pi->lyrno) {
 			for (pi->compno = pchg->compnostart, pi->picomp =
 			  &pi->picomps[pi->compno]; pi->compno < pi->numcomps &&
-			  pi->compno < JAS_CAST(int, pchg->compnoend); ++pi->compno,
+			  pi->compno < pchg->compnoend; ++pi->compno,
 			  ++pi->picomp) {
 				if (pi->rlvlno >= pi->picomp->numrlvls) {
 					continue;
@@ -222,30 +264,29 @@ skip:
 
 static int jpc_pi_nextrpcl(register jpc_pi_t *pi)
 {
-	int rlvlno;
-	jpc_pirlvl_t *pirlvl;
-	jpc_pchg_t *pchg;
-	int prchind;
-	int prcvind;
-	int *prclyrno;
-	int compno;
-	jpc_picomp_t *picomp;
-	int xstep;
-	int ystep;
+	unsigned rlvlno;
+	unsigned *prclyrno;
+	unsigned compno;
+	unsigned xstep;
+	unsigned ystep;
 	uint_fast32_t r;
 	uint_fast32_t rpx;
 	uint_fast32_t rpy;
 	uint_fast32_t trx0;
 	uint_fast32_t try0;
 
-	pchg = pi->pchg;
+	JAS_LOGDEBUGF(10, "jpc_pi_nextrpcl\n");
+
+	const jpc_pchg_t *pchg = pi->pchg;
 	if (!pi->prgvolfirst) {
 		goto skip;
 	} else {
 		pi->xstep = 0;
 		pi->ystep = 0;
+		const jpc_picomp_t *picomp;
 		for (compno = 0, picomp = pi->picomps; compno < pi->numcomps;
 		  ++compno, ++picomp) {
+			const jpc_pirlvl_t *pirlvl;
 			for (rlvlno = 0, pirlvl = picomp->pirlvls; rlvlno <
 			  picomp->numrlvls; ++rlvlno, ++pirlvl) {
 				// Check for the potential for overflow problems.
@@ -266,6 +307,12 @@ static int jpc_pi_nextrpcl(register jpc_pi_t *pi)
 		pi->prgvolfirst = 0;
 	}
 
+	if (pi->xstep == 0 || pi->ystep == 0) {
+		/* avoid division by zero */
+		jas_logerrorf("xstep and ystep must be nonzero\n");
+		return -1;
+	}
+
 	for (pi->rlvlno = pchg->rlvlnostart; pi->rlvlno < pchg->rlvlnoend &&
 	  pi->rlvlno < pi->maxrlvls; ++pi->rlvlno) {
 		for (pi->y = pi->ystart; pi->y < pi->yend; pi->y +=
@@ -274,7 +321,7 @@ static int jpc_pi_nextrpcl(register jpc_pi_t *pi)
 			  pi->xstep - (pi->x % pi->xstep)) {
 				for (pi->compno = pchg->compnostart,
 				  pi->picomp = &pi->picomps[pi->compno];
-				  pi->compno < JAS_CAST(int, pchg->compnoend) && pi->compno <
+				  pi->compno < pchg->compnoend && pi->compno <
 				  pi->numcomps; ++pi->compno, ++pi->picomp) {
 					if (pi->rlvlno >= pi->picomp->numrlvls) {
 						continue;
@@ -290,22 +337,23 @@ static int jpc_pi_nextrpcl(register jpc_pi_t *pi)
 					try0 = JPC_CEILDIV(pi->ystart, pi->picomp->vsamp << r);
 					if (((pi->x == pi->xstart &&
 					  ((trx0 << r) % (JAS_CAST(uint_fast32_t, 1) << rpx)))
-					  || !(pi->x % (JAS_CAST(uint_fast32_t, 1) << rpx))) &&
+					  || !(pi->x % (pi->picomp->hsamp << rpx))) &&
 					  ((pi->y == pi->ystart &&
 					  ((try0 << r) % (JAS_CAST(uint_fast32_t, 1) << rpy)))
-					  || !(pi->y % (JAS_CAST(uint_fast32_t, 1) << rpy)))) {
-						prchind = JPC_FLOORDIVPOW2(JPC_CEILDIV(pi->x,
+					  || !(pi->y % (pi->picomp->vsamp << rpy)))) {
+						const unsigned prchind = JPC_FLOORDIVPOW2(JPC_CEILDIV(pi->x,
 						  pi->picomp->hsamp << r), pi->pirlvl->prcwidthexpn) -
 						  JPC_FLOORDIVPOW2(trx0, pi->pirlvl->prcwidthexpn);
-						prcvind = JPC_FLOORDIVPOW2(JPC_CEILDIV(pi->y,
+						const unsigned prcvind = JPC_FLOORDIVPOW2(JPC_CEILDIV(pi->y,
 						  pi->picomp->vsamp << r), pi->pirlvl->prcheightexpn) -
 						  JPC_FLOORDIVPOW2(try0, pi->pirlvl->prcheightexpn);
 						pi->prcno = prcvind * pi->pirlvl->numhprcs + prchind;
+						if (pi->prcno >= pi->pirlvl->numprcs) {
+							return -1;
+						}
 
-						assert(pi->prcno < pi->pirlvl->numprcs);
 						for (pi->lyrno = 0; pi->lyrno <
-						  pi->numlyrs && pi->lyrno < JAS_CAST(int,
-						  pchg->lyrnoend); ++pi->lyrno) {
+						  pi->numlyrs && pi->lyrno < pchg->lyrnoend; ++pi->lyrno) {
 							prclyrno = &pi->pirlvl->prclyrnos[pi->prcno];
 							if (pi->lyrno >= *prclyrno) {
 								++(*prclyrno);
@@ -324,36 +372,35 @@ skip:
 
 static int jpc_pi_nextpcrl(register jpc_pi_t *pi)
 {
-	int rlvlno;
-	jpc_pirlvl_t *pirlvl;
-	jpc_pchg_t *pchg;
-	int prchind;
-	int prcvind;
-	int *prclyrno;
-	int compno;
-	jpc_picomp_t *picomp;
-	int xstep;
-	int ystep;
+	unsigned rlvlno;
+	unsigned *prclyrno;
+	unsigned compno;
+	unsigned xstep;
+	unsigned ystep;
 	uint_fast32_t trx0;
 	uint_fast32_t try0;
 	uint_fast32_t r;
 	uint_fast32_t rpx;
 	uint_fast32_t rpy;
 
-	pchg = pi->pchg;
+	JAS_LOGDEBUGF(10, "jpc_pi_nextpcrl\n");
+
+	const jpc_pchg_t *pchg = pi->pchg;
 	if (!pi->prgvolfirst) {
 		goto skip;
 	} else {
 		pi->xstep = 0;
 		pi->ystep = 0;
+		const jpc_picomp_t *picomp;
 		for (compno = 0, picomp = pi->picomps; compno < pi->numcomps;
 		  ++compno, ++picomp) {
+			const jpc_pirlvl_t *pirlvl;
 			for (rlvlno = 0, pirlvl = picomp->pirlvls; rlvlno <
 			  picomp->numrlvls; ++rlvlno, ++pirlvl) {
 				// Check for the potential for overflow problems.
-				if (pirlvl->prcwidthexpn + pi->picomp->numrlvls >
+				if (pirlvl->prcwidthexpn + picomp->numrlvls >
 				  JAS_UINTFAST32_NUMBITS - 2 ||
-				  pirlvl->prcheightexpn + pi->picomp->numrlvls >
+				  pirlvl->prcheightexpn + picomp->numrlvls >
 				  JAS_UINTFAST32_NUMBITS - 2) {
 					return -1;
 				}
@@ -368,13 +415,19 @@ static int jpc_pi_nextpcrl(register jpc_pi_t *pi)
 		pi->prgvolfirst = 0;
 	}
 
+	if (pi->xstep == 0 || pi->ystep == 0) {
+		/* avoid division by zero */
+		jas_logerrorf("xstep and ystep must be nonzero\n");
+		return -1;
+	}
+
 	for (pi->y = pi->ystart; pi->y < pi->yend; pi->y += pi->ystep -
 	  (pi->y % pi->ystep)) {
 		for (pi->x = pi->xstart; pi->x < pi->xend; pi->x += pi->xstep -
 		  (pi->x % pi->xstep)) {
 			for (pi->compno = pchg->compnostart, pi->picomp =
 			  &pi->picomps[pi->compno]; pi->compno < pi->numcomps
-			  && pi->compno < JAS_CAST(int, pchg->compnoend); ++pi->compno,
+			  && pi->compno < pchg->compnoend; ++pi->compno,
 			  ++pi->picomp) {
 				for (pi->rlvlno = pchg->rlvlnostart,
 				  pi->pirlvl = &pi->picomp->pirlvls[pi->rlvlno];
@@ -395,16 +448,16 @@ static int jpc_pi_nextpcrl(register jpc_pi_t *pi)
 					  ((pi->y == pi->ystart &&
 					  ((try0 << r) % (JAS_CAST(uint_fast32_t, 1) << rpy))) ||
 					  !(pi->y % (pi->picomp->vsamp << rpy)))) {
-						prchind = JPC_FLOORDIVPOW2(JPC_CEILDIV(pi->x,
+						const unsigned prchind = JPC_FLOORDIVPOW2(JPC_CEILDIV(pi->x,
 						  pi->picomp->hsamp << r), pi->pirlvl->prcwidthexpn) -
 						  JPC_FLOORDIVPOW2(trx0, pi->pirlvl->prcwidthexpn);
-						prcvind = JPC_FLOORDIVPOW2(JPC_CEILDIV(pi->y,
+						const unsigned prcvind = JPC_FLOORDIVPOW2(JPC_CEILDIV(pi->y,
 						  pi->picomp->vsamp << r), pi->pirlvl->prcheightexpn) -
 						  JPC_FLOORDIVPOW2(try0, pi->pirlvl->prcheightexpn);
 						pi->prcno = prcvind * pi->pirlvl->numhprcs + prchind;
 						assert(pi->prcno < pi->pirlvl->numprcs);
 						for (pi->lyrno = 0; pi->lyrno < pi->numlyrs &&
-						  pi->lyrno < JAS_CAST(int, pchg->lyrnoend);
+						  pi->lyrno < pchg->lyrnoend;
 						  ++pi->lyrno) {
 							prclyrno = &pi->pirlvl->prclyrnos[pi->prcno];
 							if (pi->lyrno >= *prclyrno) {
@@ -424,34 +477,45 @@ skip:
 
 static int jpc_pi_nextcprl(register jpc_pi_t *pi)
 {
-	int rlvlno;
-	jpc_pirlvl_t *pirlvl;
-	jpc_pchg_t *pchg;
-	int prchind;
-	int prcvind;
-	int *prclyrno;
+	unsigned rlvlno;
+	unsigned *prclyrno;
 	uint_fast32_t trx0;
 	uint_fast32_t try0;
 	uint_fast32_t r;
 	uint_fast32_t rpx;
 	uint_fast32_t rpy;
 
-	pchg = pi->pchg;
+	JAS_LOGDEBUGF(10, "jpc_pi_nextcprl\n");
+
+	const jpc_pchg_t *pchg = pi->pchg;
 	if (!pi->prgvolfirst) {
 		goto skip;
 	} else {
 		pi->prgvolfirst = 0;
 	}
 
+#if 0
+	/*
+	This disabled code is wrong.  xstep and ystep need not be set yet.
+	This will cause some valid code streams to fail to be decoded.
+	*/
+	if (pi->xstep == 0 || pi->ystep == 0) {
+		/* avoid later division by zero */
+		jas_logerrorf("xstep and ystep must be nonzero\n");
+		return -1;
+	}
+#endif
+
 	for (pi->compno = pchg->compnostart, pi->picomp = &pi->picomps[pi->compno];
-	  pi->compno < JAS_CAST(int, pchg->compnoend) && pi->compno < pi->numcomps;
+	  pi->compno < pchg->compnoend && pi->compno < pi->numcomps;
 	  ++pi->compno, ++pi->picomp) {
-		pirlvl = pi->picomp->pirlvls;
+		const jpc_pirlvl_t *pirlvl = pi->picomp->pirlvls;
 		// Check for the potential for overflow problems.
 		if (pirlvl->prcwidthexpn + pi->picomp->numrlvls >
 		  JAS_UINTFAST32_NUMBITS - 2 ||
 		  pirlvl->prcheightexpn + pi->picomp->numrlvls >
 		  JAS_UINTFAST32_NUMBITS - 2) {
+			jas_logerrorf("overflow detected\n");
 			return -1;
 		}
 		pi->xstep = pi->picomp->hsamp * (JAS_CAST(uint_fast32_t, 1) <<
@@ -460,12 +524,37 @@ static int jpc_pi_nextcprl(register jpc_pi_t *pi)
 		  (pirlvl->prcheightexpn + pi->picomp->numrlvls - 1));
 		for (rlvlno = 1, pirlvl = &pi->picomp->pirlvls[1];
 		  rlvlno < pi->picomp->numrlvls; ++rlvlno, ++pirlvl) {
+			/* Perform the following calculation in an overflow-safe manner,
+			setting the result to zero upon overflow:
 			pi->xstep = JAS_MIN(pi->xstep, pi->picomp->hsamp *
 			  (JAS_CAST(uint_fast32_t, 1) << (pirlvl->prcwidthexpn +
 			  pi->picomp->numrlvls - rlvlno - 1)));
+			*/
+			pi->xstep = JAS_MIN(pi->xstep, jas_safeui64_to_int(
+			  jas_safeui64_mul(
+			    jas_safeui64_from_intmax(pi->picomp->hsamp),
+			    jas_safeui64_pow2_intmax(pirlvl->prcwidthexpn +
+			      pi->picomp->numrlvls - rlvlno - 1)), 0));
+			if (!pi->xstep) {
+				jas_logerrorf("overflow in x-step calculation\n");
+				return -1;
+			}
+
+			/* Perform the following calculation in an overflow-safe manner,
+			setting the result to zero upon overflow:
 			pi->ystep = JAS_MIN(pi->ystep, pi->picomp->vsamp *
 			  (JAS_CAST(uint_fast32_t, 1) << (pirlvl->prcheightexpn +
 			  pi->picomp->numrlvls - rlvlno - 1)));
+			*/
+			pi->ystep = JAS_MIN(pi->ystep, jas_safeui64_to_int(
+			  jas_safeui64_mul(
+			    jas_safeui64_from_intmax(pi->picomp->vsamp),
+			    jas_safeui64_pow2_intmax(pirlvl->prcheightexpn +
+			      pi->picomp->numrlvls - rlvlno - 1)), 0));
+			if (!pi->ystep) {
+				jas_logerrorf("overflow in y-step calculation\n");
+				return -1;
+			}
 		}
 		for (pi->y = pi->ystart; pi->y < pi->yend;
 		  pi->y += pi->ystep - (pi->y % pi->ystep)) {
@@ -489,16 +578,16 @@ static int jpc_pi_nextcprl(register jpc_pi_t *pi)
 					  ((pi->y == pi->ystart &&
 					  ((try0 << r) % (JAS_CAST(uint_fast32_t, 1) << rpy))) ||
 					  !(pi->y % (pi->picomp->vsamp << rpy)))) {
-						prchind = JPC_FLOORDIVPOW2(JPC_CEILDIV(pi->x,
+						const unsigned prchind = JPC_FLOORDIVPOW2(JPC_CEILDIV(pi->x,
 						  pi->picomp->hsamp << r), pi->pirlvl->prcwidthexpn) -
 						  JPC_FLOORDIVPOW2(trx0, pi->pirlvl->prcwidthexpn);
-						prcvind = JPC_FLOORDIVPOW2(JPC_CEILDIV(pi->y,
+						const unsigned prcvind = JPC_FLOORDIVPOW2(JPC_CEILDIV(pi->y,
 						  pi->picomp->vsamp << r), pi->pirlvl->prcheightexpn) -
 						  JPC_FLOORDIVPOW2(try0, pi->pirlvl->prcheightexpn);
 						pi->prcno = prcvind * pi->pirlvl->numhprcs + prchind;
 						assert(pi->prcno < pi->pirlvl->numprcs);
 						for (pi->lyrno = 0; pi->lyrno < pi->numlyrs &&
-						  pi->lyrno < JAS_CAST(int, pchg->lyrnoend);
+						  pi->lyrno < pchg->lyrnoend;
 						  ++pi->lyrno) {
 							prclyrno = &pi->pirlvl->prclyrnos[pi->prcno];
 							if (pi->lyrno >= *prclyrno) {
@@ -516,30 +605,14 @@ skip:
 	return 1;
 }
 
-static void pirlvl_destroy(jpc_pirlvl_t *rlvl)
-{
-	if (rlvl->prclyrnos) {
-		jas_free(rlvl->prclyrnos);
-	}
-}
-
-static void jpc_picomp_destroy(jpc_picomp_t *picomp)
-{
-	int rlvlno;
-	jpc_pirlvl_t *pirlvl;
-	if (picomp->pirlvls) {
-		for (rlvlno = 0, pirlvl = picomp->pirlvls; rlvlno <
-		  picomp->numrlvls; ++rlvlno, ++pirlvl) {
-			pirlvl_destroy(pirlvl);
-		}
-		jas_free(picomp->pirlvls);
-	}
-}
+/******************************************************************************\
+*
+\******************************************************************************/
 
 void jpc_pi_destroy(jpc_pi_t *pi)
 {
 	jpc_picomp_t *picomp;
-	int compno;
+	unsigned compno;
 	if (pi->picomps) {
 		for (compno = 0, picomp = pi->picomps; compno < pi->numcomps;
 		  ++compno, ++picomp) {
@@ -568,11 +641,68 @@ jpc_pi_t *jpc_pi_create0()
 	return pi;
 }
 
+int jpc_pi_init(jpc_pi_t *pi)
+{
+	unsigned compno;
+	unsigned rlvlno;
+	unsigned prcno;
+	unsigned *prclyrno;
+
+	pi->prgvolfirst = 0;
+	pi->valid = 0;
+	pi->pktno = -1;
+	pi->pchgno = -1;
+	pi->pchg = 0;
+
+	const jpc_picomp_t *picomp;
+	for (compno = 0, picomp = pi->picomps; compno < pi->numcomps;
+	  ++compno, ++picomp) {
+		const jpc_pirlvl_t *pirlvl;
+		for (rlvlno = 0, pirlvl = picomp->pirlvls; rlvlno <
+		  picomp->numrlvls; ++rlvlno, ++pirlvl) {
+			for (prcno = 0, prclyrno = pirlvl->prclyrnos;
+			  prcno < pirlvl->numprcs; ++prcno, ++prclyrno) {
+				*prclyrno = 0;
+			}
+		}
+	}
+	return 0;
+}
+
 int jpc_pi_addpchg(jpc_pi_t *pi, jpc_pocpchg_t *pchg)
 {
 	return jpc_pchglist_insert(pi->pchglist, -1, pchg);
 }
 
+/* For debugging only. */
+void jpc_pi_dump(const jpc_pi_t *pi)
+{
+	jas_eprintf("numlyrs=%d\n", pi->numlyrs);
+	jas_eprintf("maxrlvls=%d\n", pi->maxrlvls);
+	jas_eprintf("numcomps=%d\n", pi->numcomps);
+	jas_eprintf("compno=%d\n", pi->compno);
+	jas_eprintf("rlvlno=%d\n", pi->rlvlno);
+	jas_eprintf("prcno=%d\n", pi->prcno);
+	jas_eprintf("lyrno=%d\n", pi->lyrno);
+	jas_eprintf("x=%d\n", pi->x);
+	jas_eprintf("y=%d\n", pi->y);
+	jas_eprintf("xstep=%d\n", pi->xstep);
+	jas_eprintf("ystep=%d\n", pi->ystep);
+	jas_eprintf("xstart=%d\n", pi->xstart);
+	jas_eprintf("ystart=%d\n", pi->ystart);
+	jas_eprintf("xend=%d\n", pi->xend);
+	jas_eprintf("yend=%d\n", pi->yend);
+	jas_eprintf("defaultpchg=%d\n", pi->defaultpchg);
+	jas_eprintf("pchgno=%d\n", pi->pchgno);
+	jas_eprintf("prgvolfirst=%d\n", pi->prgvolfirst);
+	jas_eprintf("valid=%d\n", pi->valid);
+	jas_eprintf("pktno=%d\n", pi->pktno);
+}
+
+/******************************************************************************\
+*
+\******************************************************************************/
+
 jpc_pchglist_t *jpc_pchglist_create()
 {
 	jpc_pchglist_t *pchglist;
@@ -587,14 +717,12 @@ jpc_pchglist_t *jpc_pchglist_create()
 
 int jpc_pchglist_insert(jpc_pchglist_t *pchglist, int pchgno, jpc_pchg_t *pchg)
 {
-	int i;
-	int newmaxpchgs;
 	jpc_pchg_t **newpchgs;
 	if (pchgno < 0) {
 		pchgno = pchglist->numpchgs;
 	}
 	if (pchglist->numpchgs >= pchglist->maxpchgs) {
-		newmaxpchgs = pchglist->maxpchgs + 128;
+		const unsigned newmaxpchgs = pchglist->maxpchgs + 128;
 		if (!(newpchgs = jas_realloc2(pchglist->pchgs, newmaxpchgs,
 		  sizeof(jpc_pchg_t *)))) {
 			return -1;
@@ -602,7 +730,7 @@ int jpc_pchglist_insert(jpc_pchglist_t *pchglist, int pchgno, jpc_pchg_t *pchg)
 		pchglist->maxpchgs = newmaxpchgs;
 		pchglist->pchgs = newpchgs;
 	}
-	for (i = pchglist->numpchgs; i > pchgno; --i) {
+	for (unsigned i = pchglist->numpchgs; i > (unsigned)pchgno; --i) {
 		pchglist->pchgs[i] = pchglist->pchgs[i - 1];
 	}
 	pchglist->pchgs[pchgno] = pchg;
@@ -610,20 +738,19 @@ int jpc_pchglist_insert(jpc_pchglist_t *pchglist, int pchgno, jpc_pchg_t *pchg)
 	return 0;
 }
 
-jpc_pchg_t *jpc_pchglist_remove(jpc_pchglist_t *pchglist, int pchgno)
+jpc_pchg_t *jpc_pchglist_remove(jpc_pchglist_t *pchglist, unsigned pchgno)
 {
-	int i;
 	jpc_pchg_t *pchg;
 	assert(pchgno < pchglist->numpchgs);
 	pchg = pchglist->pchgs[pchgno];
-	for (i = pchgno + 1; i < pchglist->numpchgs; ++i) {
+	for (unsigned i = pchgno + 1; i < pchglist->numpchgs; ++i) {
 		pchglist->pchgs[i - 1] = pchglist->pchgs[i];
 	}
 	--pchglist->numpchgs;
 	return pchg;
 }
 
-jpc_pchg_t *jpc_pchg_copy(jpc_pchg_t *pchg)
+jpc_pchg_t *jpc_pchg_copy(const jpc_pchg_t *pchg)
 {
 	jpc_pchg_t *newpchg;
 	if (!(newpchg = jas_malloc(sizeof(jpc_pchg_t)))) {
@@ -633,15 +760,14 @@ jpc_pchg_t *jpc_pchg_copy(jpc_pchg_t *pchg)
 	return newpchg;
 }
 
-jpc_pchglist_t *jpc_pchglist_copy(jpc_pchglist_t *pchglist)
+jpc_pchglist_t *jpc_pchglist_copy(const jpc_pchglist_t *pchglist)
 {
 	jpc_pchglist_t *newpchglist;
 	jpc_pchg_t *newpchg;
-	int pchgno;
 	if (!(newpchglist = jpc_pchglist_create())) {
 		return 0;
 	}
-	for (pchgno = 0; pchgno < pchglist->numpchgs; ++pchgno) {
+	for (unsigned pchgno = 0; pchgno < pchglist->numpchgs; ++pchgno) {
 		if (!(newpchg = jpc_pchg_copy(pchglist->pchgs[pchgno])) ||
 		  jpc_pchglist_insert(newpchglist, -1, newpchg)) {
 			jpc_pchglist_destroy(newpchglist);
@@ -653,9 +779,8 @@ jpc_pchglist_t *jpc_pchglist_copy(jpc_pchglist_t *pchglist)
 
 void jpc_pchglist_destroy(jpc_pchglist_t *pchglist)
 {
-	int pchgno;
 	if (pchglist->pchgs) {
-		for (pchgno = 0; pchgno < pchglist->numpchgs; ++pchgno) {
+		for (unsigned pchgno = 0; pchgno < pchglist->numpchgs; ++pchgno) {
 			jpc_pchg_destroy(pchglist->pchgs[pchgno]);
 		}
 		jas_free(pchglist->pchgs);
@@ -668,40 +793,12 @@ void jpc_pchg_destroy(jpc_pchg_t *pchg)
 	jas_free(pchg);
 }
 
-jpc_pchg_t *jpc_pchglist_get(jpc_pchglist_t *pchglist, int pchgno)
+const jpc_pchg_t *jpc_pchglist_get(const jpc_pchglist_t *pchglist, unsigned pchgno)
 {
 	return pchglist->pchgs[pchgno];
 }
 
-int jpc_pchglist_numpchgs(jpc_pchglist_t *pchglist)
+unsigned jpc_pchglist_numpchgs(const jpc_pchglist_t *pchglist)
 {
 	return pchglist->numpchgs;
 }
-
-int jpc_pi_init(jpc_pi_t *pi)
-{
-	int compno;
-	int rlvlno;
-	int prcno;
-	jpc_picomp_t *picomp;
-	jpc_pirlvl_t *pirlvl;
-	int *prclyrno;
-
-	pi->prgvolfirst = 0;
-	pi->valid = 0;
-	pi->pktno = -1;
-	pi->pchgno = -1;
-	pi->pchg = 0;
-
-	for (compno = 0, picomp = pi->picomps; compno < pi->numcomps;
-	  ++compno, ++picomp) {
-		for (rlvlno = 0, pirlvl = picomp->pirlvls; rlvlno <
-		  picomp->numrlvls; ++rlvlno, ++pirlvl) {
-			for (prcno = 0, prclyrno = pirlvl->prclyrnos;
-			  prcno < pirlvl->numprcs; ++prcno, ++prclyrno) {
-				*prclyrno = 0;
-			}
-		}
-	}
-	return 0;
-}
